/* 
* Author: Sujoy Biswas
* Affiliation: University of California, Santa Cruz
* History: May, 2016 
*/
#include <math.h>
#include "mex.h"
#define LSKCUBEDIM 3 // #dimension in output (LSK) tensor: ht x wd x feature
#define VERBOSE 0
#define EPSA 0.0000001
#define OPZEROGUARD 0.1e-6f
#define newzx(x, y) newzx[(y)*(M+wsize-1)+(x)]
#define newzy(x, y) newzy[(y)*(M+wsize-1)+(x)]
#define lsk(a, b, c, rmax, cmax) lsk[(c)*(rmax)*(cmax) + (b)*(rmax) + (a)]
#define K(i, j) K[(j)*wsize+(i)]
#define C11(a, b) C11[(b)*(M+nbors-1)+(a)]
#define C12(a, b) C12[(b)*(M+nbors-1)+(a)]
#define C22(a, b) C22[(b)*(M+nbors-1)+(a)]

void computeLSK(double *newzx, double *newzy, int M, int N, double *K, int wsize, 
                double *C11, double *C12, double *C22, int nbors, 
                double len, double h, double alp, double *lsk, int totalchannels)
{    
    int halfwsize, halfnbors, r, c, d, i, j, countnbors, p, q;
    double zx11, zx12, zx22, val1, val2, lambda1, lambda2, theta, 
            sval1, sval2, s1, s2, sum, part1, numer, denom, ev_1, ev_2;    
    halfwsize = (wsize - 1)/2;
    halfnbors = (nbors - 1)/2;    
    /* Handling image boundaries: reflection about horizontal axis*/
    for(r = 0; r < halfwsize; r++) {
        for(c = 0; c < N; c++) {
            /* on top */
            newzx((halfwsize-r-1), (halfwsize+c)) = newzx((r+halfwsize), (c+halfwsize));
            newzy((halfwsize-r-1), (halfwsize+c)) = newzy((r+halfwsize), (c+halfwsize));
            /* at bottom */
            newzx((halfwsize+M+r), (halfwsize+c)) = newzx((halfwsize+M-1-r), (c+halfwsize));
            newzy((halfwsize+M+r), (halfwsize+c)) = newzy((halfwsize+M-1-r), (c+halfwsize));
        }
    }    
    /* Handling image boundaries: reflection about vertical axis */
    for(r = 0; r < M+wsize-1; r++) {
        for(c = 0; c < halfwsize; c++) {
            /* on left */
            newzx(r, (halfwsize-c-1)) = newzx(r, (c+halfwsize));
            newzy(r, (halfwsize-c-1)) = newzy(r, (c+halfwsize));
            /* on right */
            newzx(r, halfwsize+N+c) = newzx(r, halfwsize+N-1-c);
            newzy(r, halfwsize+N+c) = newzy(r, halfwsize+N-1-c);
        }
    }
    /* covariance matrix computation */  
    ev_1 = 0; ev_2 = 0;
    for(r = 0; r < M; r++) {
        for(c = 0; c < N; c++) {            
            zx11 = 0; zx12 = 0; zx22 = 0;
            for(j = 0; j < wsize; j++) {
                for(i = 0; i < wsize; i++) {
                    val1 = newzx(r+i, c+j)*K(i, j);
                    zx11 += (val1*val1);
                    val2 = newzy(r+i, c+j)*K(i, j);
                    zx22 += (val2*val2);
                    zx12 += val1*val2;
                }
            }
            /* SVD computation in closed form */
            lambda1 = ((zx11 + zx22) + sqrt((zx11 - zx22)*(zx11 - zx22) + 4*zx12*zx12))/2;
            lambda2 = ((zx11 + zx22) - sqrt((zx11 - zx22)*(zx11 - zx22) + 4*zx12*zx12))/2;           
            numer = (zx11 + zx12 - lambda1);
            denom = (zx22 + zx12 - lambda1);            
            /* Thermal images have many zero-filled patches.
             Singularity guard adainst zero intensities */
            if(fabs(denom) < OPZEROGUARD) {                
                if(fabs(numer) < OPZEROGUARD) {
                   if(fabs(denom) > fabs(numer))
                    { ev_1 = 0; ev_2 = 1;}
                    else
                    { ev_1 = 1; ev_2 = 0;}
                }
                else {
                    ev_1 = 1; 
                    ev_2 = 0;
                }                
            }
            else {           
                theta = atan(-numer/denom);
                ev_1 = sin(theta); ev_2 = cos(theta);
            }            
            sval1 = sqrt(fabs(lambda1)); sval2 = sqrt(fabs(lambda2));
            part1 = pow((sval1*sval2 + EPSA) / len, alp);
            s1 = (sval1 + 1)/(sval2 + 1);
            s2 = 1.0/s1;
            C11(r+halfnbors, c+halfnbors) = part1*(s1*ev_2*ev_2+ s2*ev_1*ev_1);;
            C12(r+halfnbors, c+halfnbors) = part1*(s1-s2)*ev_1*ev_2;
            C22(r+halfnbors, c+halfnbors) = part1*(s1*ev_1*ev_1+ s2*ev_2*ev_2);
        }
    }       
    /* Handling image boundaries: reflection about horizontal axis*/
    for(r = 0; r < halfnbors; r++) {
        for(c = 0; c < N; c++) {
            /* on top */
            C11((halfnbors-r-1), (halfnbors+c)) = C11((r+halfnbors), (c+halfnbors));
            C22((halfnbors-r-1), (halfnbors+c)) = C22((r+halfnbors), (c+halfnbors));
            C12((halfnbors-r-1), (halfnbors+c)) = C12((r+halfnbors), (c+halfnbors));
            /* at bottom */
            C11((halfnbors+M+r), (halfnbors+c)) = C11((halfnbors+M-1-r), (c+halfnbors));
            C22((halfnbors+M+r), (halfnbors+c)) = C22((halfnbors+M-1-r), (c+halfnbors));
            C12((halfnbors+M+r), (halfnbors+c)) = C12((halfnbors+M-1-r), (c+halfnbors));
        }
    }    
    /* Handling image boundaries: reflection about vertical axis */
    for(r = 0; r < M+nbors-1; r++) {
        for(c = 0; c < halfnbors; c++) {
            /* on left */
            C11(r, (halfnbors-c-1)) = C11(r, (c+halfnbors));
            C22(r, (halfnbors-c-1)) = C22(r, (c+halfnbors));
            C12(r, (halfnbors-c-1)) = C12(r, (c+halfnbors));
            /* on right */
            C11(r, halfnbors+N+c) = C11(r, halfnbors+N-1-c);
            C22(r, halfnbors+N+c) = C22(r, halfnbors+N-1-c);
            C12(r, halfnbors+N+c) = C12(r, halfnbors+N-1-c);
        }
    }
    /* LSK coefficient Computations */
    for(c = halfnbors; c < N + halfnbors; c++) {
        for(r = halfnbors; r < M + halfnbors; r++) {           
            for(j=-halfnbors, countnbors=0; j<=halfnbors; j++){
                for(i=-halfnbors; i<=halfnbors; i++, countnbors++){
                    p = r + i;
                    q = c + j;
                    sum = -(C11(p, q)*i*i+2*C12(p, q)*i*j+C22(p, q)*j*j)/h;
                    lsk(r-halfnbors, c-halfnbors, countnbors, M, N) = exp(sum);
                }
            } /* all neighbors covered */           
        }
    }    
    /* LSK normalization */
    for(c = 0; c < N; c++) {
        for(r =0; r < M; r++){           
            for(d = 0, sum = 0; d < totalchannels; d++)
                sum +=  lsk(r, c, d, M, N);
            for(d = 0; d < totalchannels; d++)
                lsk(r, c, d, M, N) /= sum;
        }
    }    
}
/* The gateway function */
void mexFunction(int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{
    /* variable declarations here */
    mwSize dims[3], M, N, ndim;
    int wsize, nbors, halfwsize, r, c, totalchannels;
    double len, h, alp;
    double *zx, *zy, *K, *newzx, *newzy, *C11, *C12, *C22, *lsk;    
    /* verification of input and output arguments */
    if(nrhs != 8) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:nrhs", "Insufficient inputs: eight required.");
    }
    if(nlhs != 1) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:nlhs", "One output required.");
    }
    if( !mxIsDouble(prhs[0]) || mxIsComplex(prhs[0])) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:notDouble",
                "First input matrix must be type double.");
    }
    if( !mxIsDouble(prhs[1]) || mxIsComplex(prhs[1])) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:notDouble",
                "Second input matrix must be type double.");
    }
    if( !mxIsDouble(prhs[2]) || mxIsComplex(prhs[2])) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:notDouble",
                "Third input matrix must be type double.");
    }
    ndim = mxGetNumberOfDimensions(prhs[0]);
    if ((ndim != mxGetNumberOfDimensions(prhs[1])) || (ndim !=2)) {
        mexErrMsgIdAndTxt("MyToolbox:MCS:DimensionMismatch",
                "Number of dimensions of zx and zy must be same and 2");
    }
    M = mxGetM(prhs[0]); N = mxGetN(prhs[0]);
    if( (M != mxGetM(prhs[1])) || (N != mxGetN(prhs[1]))) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:DimensionMismatch",
                "Dimensions of zx and zy must be same");
    }
    if (!mxIsDouble(prhs[3]) || !mxIsDouble(prhs[4]) ||
            !mxIsDouble(prhs[5]) || !mxIsDouble(prhs[6]) ||
            !mxIsDouble(prhs[7])) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:Double",
                "Values for 4th to last should be double");
    }    
    /* read input data */
    zx = mxGetPr(prhs[0]);
    zy = mxGetPr(prhs[1]);
    K = mxGetPr(prhs[2]);    
    wsize = mxGetScalar(prhs[4]);    
    if ( (mxGetM(prhs[2]) != wsize) || (mxGetN(prhs[2]) != wsize)) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:Double",
                "Dimension of K does not match with wsize");
    }    
    nbors = mxGetScalar(prhs[5]);
    totalchannels = nbors*nbors;
    dims[0] = M; dims[1] = N; dims[2] = totalchannels;
    len = mxGetScalar(prhs[3]);
    h = mxGetScalar(prhs[6]);
    alp = mxGetScalar(prhs[7]);    
    /* create output data structure and fetch pointer */
    plhs[0] = mxCreateNumericArray(LSKCUBEDIM,dims,mxDOUBLE_CLASS,mxREAL);
    lsk = mxGetPr(plhs[0]);   
    if(VERBOSE) {
        printf("\n M = %d; N = %d", M, N);
        printf("\n len = %f; wsize = %d; nbors = %d; h = %f; alp = %f \n\n", len, wsize, nbors, h, alp);
    }
    /* Memory Allocation */
    if ((newzx = (double *) malloc(sizeof(double)*(M+wsize-1)*(N+wsize-1))) == NULL) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:memalloc", "Memory allocation for newzx failed");
    }
    if ((newzy = (double *) malloc(sizeof(double)*(M+wsize-1)*(N+wsize-1))) == NULL) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:memalloc", "Memory allocation for newzx failed");
    }   
    if ((C11 = (double *) malloc(sizeof(double)*(M+nbors-1)*(N+nbors-1))) == NULL) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:memalloc", "Memory allocation for newzx failed");
    }
    if ((C12 = (double *) malloc(sizeof(double)*(M+nbors-1)*(N+nbors-1))) == NULL) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:memalloc", "Memory allocation for newzx failed");
    }
    if ((C22 = (double *) malloc(sizeof(double)*(M+nbors-1)*(N+nbors-1))) == NULL) {
        mexErrMsgIdAndTxt("MyToolbox:LSK:memalloc", "Memory allocation for newzx failed");
    }   
    /* populate newzx and newzy */
    halfwsize = (wsize - 1)/2;
    for(c = 0; c < N; c++) {
        for(r = 0; r < M; r++) {
            newzx((r+halfwsize), (c+halfwsize)) = zx[c*M+r];
            newzy((r+halfwsize), (c+halfwsize)) = zy[c*M+r];
        }
    }
    /* Memory Allocation */
    computeLSK(newzx, newzy, M, N, K, wsize, C11, C12, C22, nbors, len, h, alp, lsk, totalchannels);   
    free(newzx);
    free(newzy);
    free(C11);
    free(C12);
    free(C22);
}
